import numpy as np
from numba import float64, int64

try:
    from numba.experimental import jitclass
except ImportError:
    from numba import jitclass

from ..spatialindex.hilbert_curve import distances_from_coordinates
from ..utils import _data2coord, ngjit


@ngjit
def _left_child(node):
    """
    Args:
        node: index of a binary tree node

    Returns:
        index of node's left child
    """
    return 2 * node + 1


@ngjit
def _right_child(node):
    """
    Args:
        node: index of a binary tree node

    Returns:
        index of node's right child
    """
    return 2 * node + 2


@ngjit
def _parent(node):
    """
    Args:
        node: index of a binary tree node

    Returns:
        index of node's parent
    """
    return (node - 1) // 2


@ngjit
def _distances_from_bounds(bounds, total_bounds, p):
    n = bounds.shape[1] // 2
    dim_ranges = [(total_bounds[d], total_bounds[d + n]) for d in range(n)]
    # Avoid degenerate case where there is a single unique rectangle that is a
    # single point. Increase the range by 1.0 to prevent divide by zero error
    for d in range(n):
        if dim_ranges[d][0] == dim_ranges[d][1]:
            dim_ranges[d] = (dim_ranges[d][0], dim_ranges[d][1] + 1)
    # Compute hilbert distance of the middle of each bounding box
    dim_mids = [(bounds[:, d] + bounds[:, d + n]) / 2.0 for d in range(n)]
    side_length = 2 ** p
    coords = np.zeros((bounds.shape[0], n), dtype=np.int64)
    for d in range(n):
        coords[:, d] = _data2coord(dim_mids[d], dim_ranges[d], side_length)
    hilbert_distances = distances_from_coordinates(p, coords)
    return hilbert_distances


class HilbertRtree:
    """
    This class provides a numba implementation of a read-only Hilbert R-tree
    spatial index

    See https://en.wikipedia.org/wiki/Hilbert_R-tree for more info on the Hilbert R-tree

    This implementation stores the R-tree as an array representation of a binary tree.
    See https://en.wikipedia.org/wiki/Binary_tree#Arrays for more info on the array
    representation of a binary tree.
    """

    @staticmethod
    @ngjit
    def _build_hilbert_rtree(bounds, p, page_size):
        """
        numba function to construct a Hilbert Rtree

        See HilbertRtree.__init__ for parameter descriptions
        """
        # Handle empty bounds array
        if bounds.size == 0:
            return bounds, np.zeros(0, dtype=np.int64), bounds

        # Init bounds_tree array for storing the binary tree representation
        input_size = bounds.shape[0]
        n = bounds.shape[1] // 2
        num_pages = int(np.ceil(input_size / page_size))

        tree_depth = int(np.ceil(np.log2(num_pages)))
        next_pow2 = 2 ** tree_depth
        tree_length = next_pow2 * 2 - 1
        bounds_tree = np.full((tree_length, 2 * n), np.nan)
        leaf_start = tree_length - next_pow2

        # Compute Hilbert distances for inputs
        total_bounds = ([bounds[:, d].min() for d in range(n)] +
                        [bounds[:, d + n].max() for d in range(n)])
        hilbert_distances = _distances_from_bounds(bounds, total_bounds, p)

        # Calculate indices needed to sort bounds by hilbert distance
        keys = np.argsort(hilbert_distances)

        # Populate leaves of the tree, one leaf per page. This is layer = tree_depth
        sorted_bounds = bounds[keys, :]

        for page in range(num_pages):
            start = page * page_size
            stop = start + page_size
            page_bounds = sorted_bounds[start:stop, :]
            d_mins = [np.min(page_bounds[:, d]) for d in range(n)]
            d_maxes = [np.max(page_bounds[:, d + n]) for d in range(n)]
            d_ranges = d_mins + d_maxes
            bounds_tree[leaf_start + page, :] = d_ranges

        # Populate internal layers of tree
        layer = tree_depth - 1
        start = _parent(tree_length - next_pow2)
        stop = _parent(tree_length - 1)

        while layer >= 0:
            for node in range(start, stop + 1):
                left_bounds = bounds_tree[_left_child(node), :]
                left_valid = not np.isnan(left_bounds[0])
                right_bounds = bounds_tree[_right_child(node), :]
                right_valid = not np.isnan(right_bounds[0])

                if left_valid:
                    if right_valid:
                        d_mins = [min(left_bounds[d], right_bounds[d]) for d in range(n)]
                        d_maxes = [max(left_bounds[d + n], right_bounds[d + n]) for d in range(n)]
                        d_ranges = d_mins + d_maxes
                        bounds_tree[node, :] = d_ranges
                    else:
                        bounds_tree[node, :] = left_bounds

                elif right_valid:
                    bounds_tree[node, :] = right_bounds

            # update layer, start/stop
            start = _parent(start)
            stop = _parent(stop)
            layer -= 1

        return sorted_bounds, keys, bounds_tree

    def __init__(self, bounds, p=10, page_size=512):
        """
        Construct a new HilbertRtree

        Args:
            bounds: A 2-dimensional numpy array representing a collection of
                n-dimensional bounding boxes. One row per bounding box with
                2*n columns containing the coordinates of each bounding box as follows:
                    - bounds[:, 0:n] contains the min values of the bounding boxes,
                        one column for each of the n dimensions
                    - bounds[:, n+1:2n] contains the max values of the bounding boxes,
                        one column for each of the n dimensions
            p: The Hilbert curve order parameter that determines the resolution
                of the 2D grid that data points are rounded to before computing
                their Hilbert distance. Points will be discretized into 2 ** p
                bins in both the x and y dimensions.
            page_size: Number of elements per leaf of the tree.
        """
        # Validate/coerce inputs
        if len(bounds.shape) != 2:
            raise ValueError("bounds must be a 2D array")

        if bounds.shape[1] < 2 or bounds.shape[1] % 2 != 0:
            raise ValueError(
                "The second dimension of bounds must be a multiple of 2 and at least 2"
            )

        self._page_size = max(1, page_size)  # 1 is smallest valid page size
        self._numba_rtree = None
        self._sorted_bounds, self._keys, self._bounds_tree = \
            HilbertRtree._build_hilbert_rtree(bounds.astype('float64'), p, self._page_size)

    def __getstate__(self):
        # Remove _NumbaRtree instance during serialization since jitclass instances
        # don't support it.
        state = self.__dict__
        state['_numba_rtree'] = None
        return state

    @property
    def numba_rtree(self):
        """
        Returns:
            _NumbaRtree jitclass instance that is suitable for use inside numba
            functions
        """
        if self._numba_rtree is None:
            self._numba_rtree = _NumbaRtree(
                self._sorted_bounds, self._keys, self._page_size, self._bounds_tree
            )
        return self._numba_rtree

    def intersects(self, bounds):
        """
        Compute the indices of the input bounding boxes that intersect with the
        supplied query bounds

        Args:
            bounds: An array of the form [min0, min1, ..., max0, max1, ...]
                representing the bounds to calculate intersections against

        Returns:
            1d numpy array of the indices of all of the rows in the input bounds array
            that intersect with the query bounds
        """
        bounds = tuple(float(b) for b in bounds)
        return self.numba_rtree.intersects(bounds)

    def covers_overlaps(self, bounds):
        """
        Simultaneously compute the indices of the input bounding boxes that are covered
        by the query bounds and those that overlap with the query bounds

        Args:
            bounds: An array of the form [min0, min1, ..., max0, max1, ...]
                representing the bounds to calculate intersections against

        Returns:
            Tuple of two 1d numpy arrays of indices into the input bounds array.
              * The first array contains indices of all bounding boxes that are fully
              covered by the query bounds.
              * The second array contains the indices of all bounding boxes that
              overlap with one or more edges of the query bounds.
        """
        bounds = tuple(float(b) for b in bounds)
        return self.numba_rtree.covers_overlaps(bounds)

    @property
    def empty(self):
        """
        True if the RTree was created with zero bounding boxes
        """
        return self.numba_rtree._bounds_tree.shape[0] == 0

    @property
    def total_bounds(self):
        """
        Tuple of the total bounds of all bounding boxes
        """
        if not self.empty:
            return tuple(self.numba_rtree._bounds_tree[0, :])
        else:
            return tuple((np.nan,) * self.numba_rtree._bounds_tree.shape[1])


_numbartree_spec = [
    ('_bounds', float64[:, :]),
    ('_keys', int64[:]),
    ('_page_size', int64),
    ('_bounds_tree', float64[:, :]),
]
@jitclass(_numbartree_spec)
class _NumbaRtree:
    def __init__(self, bounds, keys, page_size, bounds_tree):
        self._bounds = bounds
        self._keys = keys
        self._page_size = page_size
        self._bounds_tree = bounds_tree

    def _leaf_start(self):
        """
        Returns
            Index of the first leaf node in bounds_tree
        """
        return (self._bounds_tree.shape[0] + 1) // 2 - 1

    def _start_index(self, node):
        """
        Args
            node: Index into _bounds_tree representing a binary tree node
        Returns
            The first index into keys represented by node
        """
        leaf_start = self._leaf_start()
        while True:
            child = _left_child(node)
            if child >= self._bounds_tree.shape[0]:
                page = node - leaf_start
                return page * self._page_size
            else:
                node = child

    def _stop_index(self, node):
        """
        Args
            node: Index into _bounds_tree representing a binary tree node
        Returns
            One past the last index into keys represented by node
        """
        leaf_start = self._leaf_start()
        while True:
            child = _right_child(node)
            if child >= self._bounds_tree.shape[0]:
                page = node - leaf_start + 1
                return page * self._page_size
            else:
                node = child

    def _maybe_intersects_ranges(self, query_bounds):
        if len(query_bounds) % 2 != 0:
            raise ValueError(
                'query_bounds must an array with an even number of elements'
            )
        n = len(query_bounds) // 2
        nodes = [0]
        covered_ranges = []
        maybe_intersect_ranges = []
        # Find ranges of indices that overlap with query bounds
        while nodes:
            next_node = nodes.pop()
            node_bounds = self._bounds_tree[next_node, :]

            # Check if node's bounds are fully outside query bounds
            outside = False
            for d in range(n):
                if (query_bounds[n + d] < node_bounds[d] or
                        query_bounds[d] > node_bounds[n + d]):
                    outside = True
                    break

            if outside:
                continue

            # Check if node's bounds are fully inside query bounds
            inside = True
            for d in range(n):
                if (node_bounds[d] < query_bounds[d] or
                        node_bounds[n + d] > query_bounds[n + d]):
                    inside = False
                    break

            if inside:
                # Node's bounds are fully inside query bounds
                start = self._start_index(next_node)
                stop = self._stop_index(next_node)
                covered_ranges.append((start, stop))
            else:
                start = self._start_index(next_node)
                stop = self._stop_index(next_node)
                if stop - start <= self._page_size:
                    maybe_intersect_ranges.append((start, stop))
                else:
                    # Partial overlap of interior bounding box, recurse to children
                    nodes.extend([_right_child(next_node), _left_child(next_node)])

        return covered_ranges, maybe_intersect_ranges

    def intersects(self, query_bounds):
        """
        See HilbertRtree.intersection
        """
        if self._bounds.size == 0:
            return np.zeros(0, dtype=np.uint32)

        n = len(query_bounds) // 2
        covered_ranges, maybe_intersect_ranges = self._maybe_intersects_ranges(
            query_bounds)

        # Compute max result length
        max_len = 0
        for start, stop in covered_ranges:
            max_len += stop - start
        for start, stop in maybe_intersect_ranges:
            max_len += stop - start

        # Initialize result buffer with max length
        result = np.zeros(max_len, dtype=np.uint32)

        # populate result from covered_ranges. In this case, we have certainty that
        # all bounding boxes in each slice intersect with query_bounds.
        result_start = 0
        for start, stop in covered_ranges:
            next_slice = self._keys[start:stop]
            result[result_start:result_start + len(next_slice)] = next_slice
            result_start += len(next_slice)

        # populate result from maybe_intersect_ranges.  In this case, we need to use
        # a brute-force check to determine which bounding boxes in the slice intersect
        # with query_bounds and only include those.
        for start, stop in maybe_intersect_ranges:
            next_slice = self._keys[start:stop]
            bounds_slice = self._bounds[start:stop, :]
            outside_mask = np.zeros(bounds_slice.shape[0], dtype=np.bool_)
            for d in range(n):
                outside_mask |= (bounds_slice[:, d + n] < query_bounds[d])
                outside_mask |= (bounds_slice[:, d] > query_bounds[d + n])

            next_slice = next_slice[~outside_mask]
            result[result_start:result_start + len(next_slice)] = next_slice
            result_start += len(next_slice)

        # Return populated portion of result
        return result[:result_start]

    def covers_overlaps(self, query_bounds):
        """
        See HilbertRtree.covers_overlaps
        """
        if self._bounds.size == 0:
            return np.zeros(0, dtype=np.uint32), np.zeros(0, dtype=np.uint32)

        n = len(query_bounds) // 2
        covered_ranges, maybe_intersect_ranges = self._maybe_intersects_ranges(
            query_bounds)

        # Initialize results arrays
        max_len0 = 0
        for start, stop in covered_ranges:
            max_len0 += stop - start

        max_len1 = 0
        for start, stop in maybe_intersect_ranges:
            max_len1 += stop - start

        covers_inds = np.zeros(max_len0 + max_len1, dtype=np.uint32)
        overlaps_inds = np.zeros(max_len1, dtype=np.uint32)

        # Process covered_ranges, all of these will be added to covered_inds
        covers_start = 0
        for start, stop in covered_ranges:
            next_slice = self._keys[start:stop]
            covers_inds[covers_start:covers_start + len(next_slice)] = next_slice
            covers_start += len(next_slice)

        # Process maybe_intersect_ranges. Some of these will populate covered_inds,
        # some overlaps_inds, and some will be excluded
        overlaps_start = 0
        for start, stop in maybe_intersect_ranges:
            next_slice = self._keys[start:stop]
            bounds_slice = self._bounds[start:stop, :]

            # Check which bounds are fully outside query region
            outside_mask = np.zeros(bounds_slice.shape[0], dtype=np.bool_)
            for d in range(n):
                outside_mask |= (bounds_slice[:, d + n] < query_bounds[d])
                outside_mask |= (bounds_slice[:, d] > query_bounds[d + n])

            # Check which bounds are fully covered by query region
            covers_mask = np.ones(bounds_slice.shape[0], dtype=np.bool_)
            for d in range(n):
                covers_mask &= (bounds_slice[:, d] >= query_bounds[d])
                covers_mask &= (bounds_slice[:, d + n] <= query_bounds[d + n])

            # Update covers_inds
            covers_slice = next_slice[covers_mask]
            covers_inds[covers_start:covers_start + len(covers_slice)] = covers_slice
            covers_start += len(covers_slice)

            # Update overlaps_inds
            overlaps_slice = next_slice[~(outside_mask | covers_mask)]
            overlaps_stop = overlaps_start + len(overlaps_slice)
            overlaps_inds[overlaps_start:overlaps_stop] = overlaps_slice
            overlaps_start += len(overlaps_slice)

        return covers_inds[:covers_start], overlaps_inds[:overlaps_start]



import dask.dataframe as dd
import pytest
from hypothesis import given
from pandas.testing import assert_frame_equal, assert_series_equal

from .strategies import (
    hyp_settings,
    st_bounds,
    st_line_array,
    st_multiline_array,
    st_multipoint_array,
    st_multipolygon_array,
    st_point_array,
    st_polygon_array,
)
from spatialpandas import GeoDataFrame, GeoSeries
from spatialpandas.geometry import (
    LineArray,
    MultiLineArray,
    MultiPointArray,
    MultiPolygonArray,
    PointArray,
    PolygonArray,
)


def get_slices(v0, v1):
    return [slice(v0, v1), slice(None, v1), slice(v0, None), slice(None, None)]


@pytest.mark.slow
@given(st_point_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_point_cx_selection(gp_point, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = PointArray.from_geopandas(gp_point.cx[xslice, yslice])
            result = PointArray.from_geopandas(gp_point).cx[xslice, yslice]
            assert all(expected == result)


@pytest.mark.slow
@given(st_multipoint_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multipoint_cx_selection(gp_multipoint, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = MultiPointArray.from_geopandas(gp_multipoint.cx[xslice, yslice])
            result = MultiPointArray.from_geopandas(gp_multipoint).cx[xslice, yslice]
            assert all(expected == result)


@pytest.mark.slow
@given(st_line_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_line_cx_selection(gp_line, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = LineArray.from_geopandas(gp_line.cx[xslice, yslice])
            result = LineArray.from_geopandas(gp_line).cx[xslice, yslice]
            assert all(expected == result)


@pytest.mark.slow
@given(st_multiline_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multiline_cx_selection(gp_multiline, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = MultiLineArray.from_geopandas(gp_multiline.cx[xslice, yslice])
            result = MultiLineArray.from_geopandas(gp_multiline).cx[xslice, yslice]
            assert all(expected == result)


@pytest.mark.slow
@given(
    st_polygon_array(min_size=1, geoseries=True),
    st_bounds(
        x_min=-150, x_max=150, y_min=-150, y_max=150, orient=True
    )
)
@hyp_settings
def test_polygon_cx_selection(gp_polygon, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = PolygonArray.from_geopandas(gp_polygon.cx[xslice, yslice])
            result = PolygonArray.from_geopandas(gp_polygon).cx[xslice, yslice]
            assert all(expected == result)


@pytest.mark.slow
@given(
    st_multipolygon_array(min_size=1, geoseries=True),
    st_bounds(
        x_min=-150, x_max=150, y_min=-150, y_max=150, orient=True
    )
)
@hyp_settings
def test_multipolygon_cx_selection(gp_multipolygon, rect):
    x0, y0, x1, y1 = rect
    for xslice in get_slices(x0, x1):
        for yslice in get_slices(y0, y1):
            expected = MultiPolygonArray.from_geopandas(
                gp_multipolygon.cx[xslice, yslice]
            )
            result = MultiPolygonArray.from_geopandas(
                gp_multipolygon
            ).cx[xslice, yslice]
            assert all(expected == result)


@given(st_multipoint_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multipoint_cx_series_selection(gp_multipoint, rect):
    x0, y0, x1, y1 = rect
    expected = GeoSeries(gp_multipoint.cx[x0:x1, y0:y1], dtype='multipoint')

    sp_multipoint = GeoSeries(gp_multipoint)
    result = sp_multipoint.cx[x0:x1, y0:y1]
    assert_series_equal(expected, result, obj='GeoSeries')


@pytest.mark.slow
@given(st_multipoint_array(min_size=6, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multipoint_cx_series_selection_dask(gp_multipoint, rect):
    x0, y0, x1, y1 = rect
    expected = GeoSeries(gp_multipoint.cx[x0:x1, y0:y1], dtype='multipoint')

    sp_multipoint = dd.from_pandas(GeoSeries(gp_multipoint), npartitions=3)
    result = sp_multipoint.cx[x0:x1, y0:y1].compute()
    assert_series_equal(expected, result, obj='GeoSeries')


@given(st_multipoint_array(min_size=1, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multipoint_cx_frame_selection(gp_multipoint, rect):
    x0, y0, x1, y1 = rect
    expected = GeoDataFrame(
        GeoSeries(gp_multipoint.cx[x0:x1, y0:y1], dtype='multipoint')
    )

    sp_multipoint = GeoSeries(gp_multipoint).to_frame()
    result = sp_multipoint.cx[x0:x1, y0:y1]
    assert_frame_equal(expected, result, obj='GeoDataFrame')


@pytest.mark.slow
@given(st_multipoint_array(min_size=6, geoseries=True), st_bounds(orient=True))
@hyp_settings
def test_multipoint_cx_frame_selection_dask(gp_multipoint, rect):
    x0, y0, x1, y1 = rect
    expected = GeoDataFrame(
        GeoSeries(gp_multipoint.cx[x0:x1, y0:y1], dtype='multipoint')
    )

    sp_multipoint = dd.from_pandas(GeoSeries(gp_multipoint).to_frame(), npartitions=3)
    result = sp_multipoint.cx[x0:x1, y0:y1].compute()
    assert_frame_equal(expected, result, obj='GeoDataFrame')



import dask
import dask.dataframe as dd
import hypothesis.strategies as hs
import numpy as np
import pandas as pd
from pathlib import Path
import pytest
from hypothesis import HealthCheck, Phase, Verbosity, given, settings

from .geometry.strategies import (
    st_bounds,
    st_multiline_array,
    st_multipoint_array,
    st_point_array,
)
from spatialpandas import GeoDataFrame, GeoSeries
from spatialpandas.dask import DaskGeoDataFrame
from spatialpandas.io import read_parquet, read_parquet_dask, to_parquet

dask.config.set(scheduler="single-threaded")

hyp_settings = settings(
    deadline=None,
    max_examples=100,
    suppress_health_check=[HealthCheck.too_slow],
)


@given(
    gp_point=st_point_array(min_size=1, geoseries=True),
    gp_multipoint=st_multipoint_array(min_size=1, geoseries=True),
    gp_multiline=st_multiline_array(min_size=1, geoseries=True),
)
@hyp_settings
def test_parquet(gp_point, gp_multipoint, gp_multiline, tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe
    n = min(len(gp_multipoint), len(gp_multiline))
    df = GeoDataFrame({
        'point': GeoSeries(gp_point[:n]),
        'multipoint': GeoSeries(gp_multipoint[:n]),
        'multiline': GeoSeries(gp_multiline[:n]),
        'a': list(range(n))
    })

    df.index.name = 'range_idx'

    path = tmp_path / 'df.parq'
    to_parquet(df, path)
    df_read = read_parquet(str(path), columns=['point', 'multipoint', 'multiline', 'a'])
    assert isinstance(df_read, GeoDataFrame)
    pd.testing.assert_frame_equal(df, df_read)
    assert df_read.index.name == df.index.name


@given(
    gp_point=st_point_array(min_size=1, geoseries=True),
    gp_multipoint=st_multipoint_array(min_size=1, geoseries=True),
    gp_multiline=st_multiline_array(min_size=1, geoseries=True),
)
@hyp_settings
def test_parquet_columns(gp_point, gp_multipoint, gp_multiline,
                         tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe
    n = min(len(gp_multipoint), len(gp_multiline))
    df = GeoDataFrame({
        'point': GeoSeries(gp_point[:n]),
        'multipoint': GeoSeries(gp_multipoint[:n]),
        'multiline': GeoSeries(gp_multiline[:n]),
        'a': list(range(n))
    })

    path = tmp_path / 'df.parq'
    to_parquet(df, path)
    columns = ['a', 'multiline']
    df_read = read_parquet(str(path), columns=columns)
    assert isinstance(df_read, GeoDataFrame)
    pd.testing.assert_frame_equal(df[columns], df_read)


@given(
    gp_multipoint=st_multipoint_array(min_size=1, geoseries=True),
    gp_multiline=st_multiline_array(min_size=1, geoseries=True),
)
@hyp_settings
def test_parquet_dask(gp_multipoint, gp_multiline, tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe
    n = min(len(gp_multipoint), len(gp_multiline))
    df = GeoDataFrame({
        'points': GeoSeries(gp_multipoint[:n]),
        'lines': GeoSeries(gp_multiline[:n]),
        'a': list(range(n))
    })
    ddf = dd.from_pandas(df, npartitions=3)

    path = tmp_path / 'ddf.parq'
    ddf.to_parquet(str(path))
    ddf_read = read_parquet_dask(str(path))

    # Check type
    assert isinstance(ddf_read, DaskGeoDataFrame)

    # Check that partition bounds were loaded
    nonempty = np.nonzero(
        np.asarray(ddf.map_partitions(len).compute() > 0)
    )[0]
    assert set(ddf_read._partition_bounds) == {'points', 'lines'}
    expected_partition_bounds = (
        ddf['points'].partition_bounds.iloc[nonempty].reset_index(drop=True)
    )
    expected_partition_bounds.index.name = 'partition'

    pd.testing.assert_frame_equal(
        expected_partition_bounds,
        ddf_read._partition_bounds['points'],
    )

    expected_partition_bounds = (
        ddf['lines'].partition_bounds.iloc[nonempty].reset_index(drop=True)
    )
    expected_partition_bounds.index.name = 'partition'
    pd.testing.assert_frame_equal(
        expected_partition_bounds,
        ddf_read._partition_bounds['lines'],
    )

    assert ddf_read.geometry.name == 'points'


@given(
    gp_multipoint=st_multipoint_array(min_size=10, max_size=40, geoseries=True),
    gp_multiline=st_multiline_array(min_size=10, max_size=40, geoseries=True),
)
@settings(deadline=None, max_examples=30)
def test_pack_partitions(gp_multipoint, gp_multiline):
    # Build dataframe
    n = min(len(gp_multipoint), len(gp_multiline))
    df = GeoDataFrame({
        'points': GeoSeries(gp_multipoint[:n]),
        'lines': GeoSeries(gp_multiline[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf = dd.from_pandas(df, npartitions=3)

    # Pack partitions
    ddf_packed = ddf.pack_partitions(npartitions=4)

    # Check the number of partitions
    assert ddf_packed.npartitions == 4

    # Check that rows are now sorted in order of hilbert distance
    total_bounds = df.lines.total_bounds
    hilbert_distances = ddf_packed.lines.map_partitions(
        lambda s: s.hilbert_distance(total_bounds=total_bounds)
    ).compute().values

    # Compute expected total_bounds
    expected_distances = np.sort(
        df.lines.hilbert_distance(total_bounds=total_bounds).values
    )

    np.testing.assert_equal(expected_distances, hilbert_distances)


@pytest.mark.slow
@given(
    gp_multipoint=st_multipoint_array(min_size=60, max_size=100, geoseries=True),
    gp_multiline=st_multiline_array(min_size=60, max_size=100, geoseries=True),
    use_temp_format=hs.booleans()
)
@settings(
    deadline=None,
    max_examples=30,
    suppress_health_check=[HealthCheck.too_slow],
    phases=[
        Phase.explicit,
        Phase.reuse,
        Phase.generate,
        Phase.target
    ],
    verbosity=Verbosity.verbose,
)
def test_pack_partitions_to_parquet(gp_multipoint, gp_multiline,
                                    use_temp_format, tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe
    n = min(len(gp_multipoint), len(gp_multiline))
    df = GeoDataFrame({
        'points': GeoSeries(gp_multipoint[:n]),
        'lines': GeoSeries(gp_multiline[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf = dd.from_pandas(df, npartitions=3)

    path = tmp_path / 'ddf.parq'
    if use_temp_format:
        (tmp_path / 'scratch').mkdir(parents=True, exist_ok=True)
        tempdir_format = str(tmp_path / 'scratch' / 'part-{uuid}-{partition:03d}')
    else:
        tempdir_format = None

    _retry_args = dict(
        wait_exponential_multiplier=10,
        wait_exponential_max=20000,
        stop_max_attempt_number=4
    )

    ddf_packed = ddf.pack_partitions_to_parquet(
        str(path),
        npartitions=12,
        tempdir_format=tempdir_format,
        _retry_args=_retry_args,
    )

    # Check the number of partitions (< 4 can happen in the case of empty partitions)
    assert ddf_packed.npartitions <= 12

    # Check that rows are now sorted in order of hilbert distance
    total_bounds = df.lines.total_bounds
    hilbert_distances = ddf_packed.lines.map_partitions(
        lambda s: s.hilbert_distance(total_bounds=total_bounds)
    ).compute().values

    # Compute expected total_bounds
    expected_distances = np.sort(
        df.lines.hilbert_distance(total_bounds=total_bounds).values
    )

    np.testing.assert_equal(expected_distances, hilbert_distances)
    assert ddf_packed.geometry.name == 'points'

    # Read columns
    columns = ['a', 'lines']
    ddf_read_cols = read_parquet_dask(path, columns=columns)
    pd.testing.assert_frame_equal(
        ddf_read_cols.compute(), ddf_packed[columns].compute()
    )


@pytest.mark.slow
@given(
    gp_multipoint1=st_multipoint_array(min_size=10, max_size=40, geoseries=True),
    gp_multiline1=st_multiline_array(min_size=10, max_size=40, geoseries=True),
    gp_multipoint2=st_multipoint_array(min_size=10, max_size=40, geoseries=True),
    gp_multiline2=st_multiline_array(min_size=10, max_size=40, geoseries=True),
)
@settings(deadline=None, max_examples=30, suppress_health_check=[HealthCheck.too_slow])
def test_pack_partitions_to_parquet_glob(gp_multipoint1, gp_multiline1,
                                         gp_multipoint2, gp_multiline2,
                                         tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe1
    n = min(len(gp_multipoint1), len(gp_multiline1))
    df1 = GeoDataFrame({
        'points': GeoSeries(gp_multipoint1[:n]),
        'lines': GeoSeries(gp_multiline1[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf1 = dd.from_pandas(df1, npartitions=3)
    path1 = tmp_path / 'ddf1.parq'
    ddf_packed1 = ddf1.pack_partitions_to_parquet(str(path1), npartitions=3)

    # Build dataframe2
    n = min(len(gp_multipoint2), len(gp_multiline2))
    df2 = GeoDataFrame({
        'points': GeoSeries(gp_multipoint2[:n]),
        'lines': GeoSeries(gp_multiline2[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf2 = dd.from_pandas(df2, npartitions=3)
    path2 = tmp_path / 'ddf2.parq'
    ddf_packed2 = ddf2.pack_partitions_to_parquet(str(path2), npartitions=4)

    # Load both packed datasets with glob
    ddf_globbed = read_parquet_dask(tmp_path / "ddf*.parq", geometry="lines")

    # Check the number of partitions (< 7 can happen in the case of empty partitions)
    assert ddf_globbed.npartitions <= 7

    # Check contents
    expected_df = pd.concat([ddf_packed1.compute(), ddf_packed2.compute()])
    df_globbed = ddf_globbed.compute()
    pd.testing.assert_frame_equal(df_globbed, expected_df)

    # Check partition bounds
    expected_bounds = {
        'points': pd.concat([
            ddf_packed1._partition_bounds['points'],
            ddf_packed2._partition_bounds['points'],
        ]).reset_index(drop=True),
        'lines': pd.concat([
            ddf_packed1._partition_bounds['lines'],
            ddf_packed2._partition_bounds['lines'],
        ]).reset_index(drop=True),
    }
    expected_bounds['points'].index.name = 'partition'
    expected_bounds['lines'].index.name = 'partition'
    pd.testing.assert_frame_equal(
        expected_bounds['points'], ddf_globbed._partition_bounds['points']
    )

    pd.testing.assert_frame_equal(
        expected_bounds['lines'], ddf_globbed._partition_bounds['lines']
    )

    assert ddf_globbed.geometry.name == 'lines'


@pytest.mark.slow
@given(
    gp_multipoint1=st_multipoint_array(min_size=10, max_size=40, geoseries=True),
    gp_multiline1=st_multiline_array(min_size=10, max_size=40, geoseries=True),
    gp_multipoint2=st_multipoint_array(min_size=10, max_size=40, geoseries=True),
    gp_multiline2=st_multiline_array(min_size=10, max_size=40, geoseries=True),
    bounds=st_bounds(),
)
@settings(deadline=None, max_examples=30, suppress_health_check=[HealthCheck.too_slow])
def test_pack_partitions_to_parquet_list_bounds(
        gp_multipoint1, gp_multiline1,
        gp_multipoint2, gp_multiline2,
        bounds, tmp_path_factory,
):
    tmp_path = tmp_path_factory.mktemp("spatialpandas", numbered=True)
    # Build dataframe1
    n = min(len(gp_multipoint1), len(gp_multiline1))
    df1 = GeoDataFrame({
        'points': GeoSeries(gp_multipoint1[:n]),
        'lines': GeoSeries(gp_multiline1[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf1 = dd.from_pandas(df1, npartitions=3)
    path1 = tmp_path / 'ddf1.parq'
    ddf_packed1 = ddf1.pack_partitions_to_parquet(str(path1), npartitions=3)

    # Build dataframe2
    n = min(len(gp_multipoint2), len(gp_multiline2))
    df2 = GeoDataFrame({
        'points': GeoSeries(gp_multipoint2[:n]),
        'lines': GeoSeries(gp_multiline2[:n]),
        'a': list(range(n))
    }).set_geometry('lines')
    ddf2 = dd.from_pandas(df2, npartitions=3)
    path2 = tmp_path / 'ddf2.parq'
    ddf_packed2 = ddf2.pack_partitions_to_parquet(str(path2), npartitions=4)

    # Load both packed datasets with glob
    ddf_read = read_parquet_dask(
        [str(tmp_path / "ddf1.parq"), str(tmp_path / "ddf2.parq")],
        geometry="points", bounds=bounds
    )

    # Check the number of partitions (< 7 can happen in the case of empty partitions)
    assert ddf_read.npartitions <= 7

    # Check contents
    xslice = slice(bounds[0], bounds[2])
    yslice = slice(bounds[1], bounds[3])
    expected_df = pd.concat([
        ddf_packed1.cx_partitions[xslice, yslice].compute(),
        ddf_packed2.cx_partitions[xslice, yslice].compute()
    ])
    df_read = ddf_read.compute()
    pd.testing.assert_frame_equal(df_read, expected_df)

    # Compute expected partition bounds
    points_bounds = pd.concat([
        ddf_packed1._partition_bounds['points'],
        ddf_packed2._partition_bounds['points'],
    ]).reset_index(drop=True)

    x0, y0, x1, y1 = bounds
    x0, x1 = (x0, x1) if x0 <= x1 else (x1, x0)
    y0, y1 = (y0, y1) if y0 <= y1 else (y1, y0)
    partition_inds = ~(
        (points_bounds.x1 < x0) |
        (points_bounds.y1 < y0) |
        (points_bounds.x0 > x1) |
        (points_bounds.y0 > y1)
    )
    points_bounds = points_bounds[partition_inds].reset_index(drop=True)

    lines_bounds = pd.concat([
        ddf_packed1._partition_bounds['lines'],
        ddf_packed2._partition_bounds['lines'],
    ]).reset_index(drop=True)[partition_inds].reset_index(drop=True)
    points_bounds.index.name = 'partition'
    lines_bounds.index.name = 'partition'

    # Check partition bounds
    pd.testing.assert_frame_equal(
        points_bounds, ddf_read._partition_bounds['points']
    )

    pd.testing.assert_frame_equal(
        lines_bounds, ddf_read._partition_bounds['lines']
    )

    # Check active geometry column
    assert ddf_read.geometry.name == 'points'


@pytest.mark.parametrize("filename", ["serial_5.0.0.parq", "serial_8.0.0.parq"])
def test_read_parquet(filename):
    path = Path(__file__).parent.joinpath("test_data", filename)
    df = read_parquet(str(path))

    assert isinstance(df, GeoDataFrame)
    assert all(df.columns == ["multiline", "a"])
    assert all(df.a == np.arange(5))
    assert df.geometry.name == "multiline"


@pytest.mark.parametrize(
    "directory, repartitioned",
    [("dask_5.0.0.parq", False), ("dask_repart_5.0.0.parq", True),
     ("dask_8.0.0.parq", False), ("dask_repart_8.0.0.parq", True)])
def test_read_parquet_dask(directory, repartitioned):
    path = Path(__file__).parent.joinpath("test_data", directory)
    ddf = read_parquet_dask(str(path))

    assert isinstance(ddf, DaskGeoDataFrame)
    assert all(ddf.columns == ["multiline", "a"])
    assert ddf.geometry.name == "multiline"
    assert ddf.npartitions == 2

    if repartitioned:
        assert all(sorted(ddf.a.compute().values) == np.arange(5))
        assert ddf.index.name == "hilbert_distance"
    else:
        assert all(ddf.a.compute() == np.arange(5))

    # Check metadata partition bounds equal the individual partition bounds.
    partition_bounds = ddf._partition_bounds
    assert list(partition_bounds) == ["multiline"]
    assert partition_bounds["multiline"].index.name == "partition"
    assert ddf["multiline"].partition_bounds.index.name == "partition"
    assert all(partition_bounds["multiline"] == ddf["multiline"].partition_bounds)



from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Union

import fsspec

PathType = Union[PathLike, str, Path]


def validate_coerce_filesystem(
    path: PathType,
    filesystem: Optional[Union[str, fsspec.AbstractFileSystem]] = None,
    storage_options: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> fsspec.AbstractFileSystem:
    """
    Validate filesystem argument and return an fsspec file system object

    Args:
        path: Path as a string
        filesystem: Optional fsspec filesystem object to use to open the file. If not
            provided, filesystem type is inferred from path
        storage_options: Key/value pairs to be passed on to the file-system backend, if any.

    Returns:
        fsspec file system
    """
    if isinstance(filesystem, fsspec.AbstractFileSystem):
        return filesystem
    fsspec_opts = kwargs.copy()
    if storage_options:
        fsspec_opts.update(storage_options)
    if filesystem is None:
        return fsspec.open(path, **fsspec_opts).fs
    else:
        try:
            return fsspec.filesystem(filesystem, **fsspec_opts)
        except ValueError:
            raise ValueError(
                "Received invalid filesystem value with type: {typ}".format(
                    typ=type(filesystem)
                )
            )


def _maybe_prepend_protocol(
    paths: Iterable[PathType],
    filesystem: fsspec.AbstractFileSystem,
) -> Iterable[PathType]:
    protocol = filesystem.protocol if isinstance(
        filesystem.protocol, str) else filesystem.protocol[0]
    if protocol not in ("file", "abstract"):
        # Add back prefix (e.g. s3://)
        p = f"{protocol}://"
        paths = [f"{p}{_}" if not _.startswith(p) else _ for _ in paths]
    return paths



import re
from collections.abc import Container, Iterable
from numbers import Integral

import numpy as np
import pandas as pd
import pyarrow as pa
from pandas.api.extensions import ExtensionArray, ExtensionDtype
from pandas.api.types import is_array_like

from .._optional_imports import gp, sg
from ..spatialindex import HilbertRtree
from ..spatialindex.rtree import _distances_from_bounds
from ..utils import ngjit


def _unwrap_geometry(a, element_dtype):
    try:
        if np.isscalar(a) and np.isnan(a):
            # replace top-level nana with None
            return None
    except (TypeError, ValueError):
        # Not nan, continue
        pass
    if isinstance(a, Geometry):
        return a.data.as_py()
    elif sg and isinstance(a, sg.base.BaseGeometry):
        return element_dtype._shapely_to_coordinates(a)
    else:
        return a


class GeometryDtype(ExtensionDtype):
    _geometry_name = 'geometry'
    base = np.dtype('O')
    _metadata = ('subtype',)
    na_value = np.nan

    def __from_arrow__(self, data):
        return self.construct_array_type()(data, dtype=self)

    @classmethod
    def _arrow_element_type_from_numpy_subtype(cls, subtype):
        raise NotImplementedError

    @classmethod
    def construct_array_type(cls, *args):
        return GeometryArray

    @classmethod
    def _parse_subtype(cls, dtype_string):
        # Be case insensitive
        dtype_string = dtype_string.lower()
        subtype_re = re.compile('^' + cls._geometry_name + r"\[(?P<subtype>\w+)\]$")

        match = subtype_re.match(dtype_string)
        if match:
            subtype_string = match.groupdict()['subtype']
        elif dtype_string == cls._geometry_name.lower():
            subtype_string = 'float64'
        else:
            raise ValueError("Cannot parse {dtype_string}".format(
                dtype_string=dtype_string))

        return subtype_string

    @classmethod
    def construct_from_string(cls, string):
        # lowercase string
        try:
            string = string.lower()
            if not isinstance(string, str):
                raise AttributeError
        except AttributeError:
            raise TypeError(
                "'construct_from_string' expects a string, got {typ}".format(
                    typ=type(string)))

        msg = "Cannot construct a '%s' from '{}'" % cls.__name__
        if string.startswith(cls._geometry_name.lower()):
            # Extract subtype
            try:
                subtype_string = cls._parse_subtype(string)
                return cls(subtype_string)
            except Exception:
                raise TypeError(msg.format(string))
        else:
            raise TypeError(msg.format(string))

    def __init__(self, subtype):
        if isinstance(subtype, GeometryDtype):
            self.subtype = subtype.subtype
        else:
            self.subtype = np.dtype(subtype)

        # Validate the subtype is numeric
        if self.subtype.kind not in ('i', 'u', 'f'):
            raise ValueError("Received non-numeric type of kind '{}'".format(self.kind))

        array_type = self.construct_array_type()
        self.arrow_dtype = array_type._arrow_type_from_numpy_element_dtype(subtype)

    def __hash__(self):
        return hash((self.__class__, self.arrow_dtype))

    def __str__(self):
        return "{}[{}]".format(self._geometry_name, str(self.subtype.name))

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__,  str(self.subtype.name))

    @property
    def type(self):
        # type: () -> type
        """The scalar type for the array, e.g. ``int``.
        It's expected ``ExtensionArray[item]`` returns an instance
        of ``ExtensionDtype.type`` for scalar ``item``.
        """
        return Geometry

    @property
    def name(self):
        # type: () -> str
        """A string identifying the data type.
        Will be used for display in, e.g. ``Series.dtype``
        """
        return str(self)


class Geometry:
    def __init__(self, data, dtype=None):
        if isinstance(data, pa.Scalar):
            # Use arrow Scalar as is
            self.data = data
        else:
            # Convert to arrow Scalar
            self.data = pa.array([data])[0]

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.data.as_py())

    def __hash__(self):
        return hash((self.__class__, np.array(self.data.as_py()).tobytes()))

    def __eq__(self, other):
        if isinstance(other, Container):
            return other == self
        if type(other) is not type(self):
            return False
        return self.data == other.data

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        raise NotImplementedError()

    @classmethod
    def from_shapely(cls, shape):
        """
        Build a spatialpandas geometry object from a shapely shape

        Args:
            shape: A shapely shape

        Returns:
            spatialpandas geometry object with type of the calling class
        """
        shape_parts = cls._shapely_to_coordinates(shape)
        return cls(shape_parts)

    def intersects_bounds(self, bounds):
        raise NotImplementedError()

    def intersects(self, shape):
        raise NotImplementedError(
            "intersects not yet implemented for %s objects" % type(self).__name__
        )


class GeometryArray(ExtensionArray):
    _element_type = Geometry
    _can_hold_na = True

    # Import / export methods
    @classmethod
    def from_geopandas(cls, ga):
        """
        Build a spatialpandas geometry array from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries to import

        Returns:
            spatialpandas geometry array with type of the calling class
        """
        if cls is GeometryArray:
            raise ValueError(
                "from_geopandas must be called on a subclass of GeometryArray"
            )
        return cls([
            cls._element_type._shapely_to_coordinates(shape)
            if shape is not None else None
            for shape in ga
        ])

    def to_geopandas(self):
        """
        Convert a spatialpandas geometry array into a geopandas GeometryArray

        Returns:
            geopandas GeometryArray
        """
        from geopandas.array import from_shapely
        return from_shapely([el.to_shapely() for el in self])

    # Constructor
    def __init__(self, array, dtype=None, copy=None):
        # Choose default dtype for empty arrays
        try:
            if len(array) == 0 and dtype is None:
                dtype = 'float64'
        except:
            # len failed
            pass

        # See if we can determine arrow array type
        if isinstance(dtype, GeometryDtype):
            # Use arrow type as-is
            arrow_dtype = dtype.arrow_dtype
        elif isinstance(dtype, pa.DataType):
            arrow_dtype = dtype
        elif dtype is not None and dtype != np.dtype('object'):
            # Scalar element dtype
            arrow_dtype = self._arrow_type_from_numpy_element_dtype(dtype)
        else:
            # Let arrow infer type
            arrow_dtype = None

        # Unwrap GeometryList elements to numpy arrays
        if is_array_like(array) or isinstance(array, list):
            array = [_unwrap_geometry(el, self._element_type) for el in array]
            array = pa.array(array, type=arrow_dtype)
        elif isinstance(array, pa.Array):
            # Nothing to do
            pass
        elif isinstance(array, pa.ChunkedArray):
            array = pa.concat_arrays(array.chunks)
        else:
            raise ValueError(
                "Unsupported type passed for {}: {}".format(
                    self.__class__.__name__, type(array)
                )
            )

        # Save off pyarrow array
        self.data = array

        # Compute types
        np_type = self._numpy_element_dtype_from_arrow_type(self.data.type)
        self._numpy_element_type = np.dtype(np_type)
        self._dtype = self._dtype_class(np_type)

        # Initialize backing property for spatial index
        self._sindex = None

    @classmethod
    def _arrow_type_from_numpy_element_dtype(cls, dtype):
        raise NotImplementedError

    def _numpy_element_dtype_from_arrow_type(self, pyarrow_type):
        raise NotImplementedError

    @property
    def _dtype_class(self):
        return GeometryDtype

    @property
    def numpy_dtype(self):
        return self._numpy_element_type

    # Arrow conversion
    def __arrow_array__(self, type=None):
        return self.data

    # ExtensionArray methods
    @property
    def dtype(self):
        return self._dtype

    def astype(self, dtype, copy=True):
        if self.dtype == dtype:
            return self.copy() if copy else self

        if dtype == np.dtype('object'):
            return np.array(self, dtype='object')

        if isinstance(dtype, GeometryDtype):
            dtype = dtype.arrow_dtype.to_pandas_dtype()
        elif isinstance(dtype, pa.DataType):
            dtype = dtype.to_pandas_dtype()
        else:
            dtype = np.dtype(dtype)

        return self.__class__(np.asarray(self.data), dtype=dtype)

    astype.__doc__ = ExtensionArray.astype.__doc__

    @property
    def nbytes(self):
        size = 0
        for buf in self.data.buffers():
            if buf is not None:
                size += buf.size
        return size

    def isna(self):
        return _extract_isnull_bytemap(self.data)

    isna.__doc__ = ExtensionArray.isna.__doc__

    def copy(self):
        return type(self)(self.data, self.dtype)

    copy.__doc__ = ExtensionArray.copy.__doc__

    def __eq__(self, other):
        if type(other) is type(self):
            if len(other) != len(self):
                raise ValueError("""
Cannot check equality of {typ} instances of unequal length
    len(ra1) == {len_a1}
    len(ra2) == {len_a2}""".format(
                    typ=type(self).__name__,
                    len_a1=len(self),
                    len_a2=len(other)))
            result = np.zeros(len(self), dtype=np.bool_)
            for i in range(len(self)):
                result[i] = self[i] == other[i]
            return result
        if isinstance(other, (self.dtype.type, type(None))):
            result = np.zeros(len(self), dtype=np.bool_)
            for i in range(len(self)):
                result[i] = self[i] == other
            return result
        raise ValueError("""
Cannot check equality of {typ} of length {a_len} with:
    {other}""".format(typ=type(self).__name__, a_len=len(self), other=repr(other)))

    def __contains__(self, item) -> bool:
        raise NotImplementedError

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        err_msg = ("Only integers, slices and integer or boolean"
                   "arrays are valid indices.")
        if isinstance(item, tuple) and len(item) == 2:
            if item[0] is Ellipsis:
                item = item[1]
            elif item[1] is Ellipsis:
                item = item[0]

        if isinstance(item, Integral):
            item = int(item)
            if item < -len(self) or item >= len(self):
                raise IndexError("{item} is out of bounds".format(item=item))
            else:
                # Convert negative item index
                if item < 0:
                    item += len(self)

                value = self.data[item].as_py()
                if value is not None:
                    return self._element_type(value, self.numpy_dtype)
                else:
                    return None
        elif isinstance(item, slice):
            if item.step is None or item.step == 1:
                # pyarrow only supports slice with step of 1
                return self.__class__(self.data[item], dtype=self.dtype)
            else:
                selected_indices = np.arange(len(self))[item]
                return self.take(selected_indices, allow_fill=False)
        elif isinstance(item, Iterable):
            if isinstance(item, (np.ndarray, ExtensionArray)):
                # Leave numpy and pandas arrays alone
                kind = item.dtype.kind
            else:
                item = pd.array(item)
                kind = item.dtype.kind

            if len(item) == 0:
                return self.take([], allow_fill=False)
            elif kind == 'b':
                # Check mask length is compatible
                if len(item) != len(self):
                    raise IndexError(
                        "Boolean index has wrong length: {} instead of {}"
                        .format(len(item), len(self))
                    )

                # check for NA values
                if any(pd.isna(item)):
                    raise ValueError(
                        "Cannot mask with a boolean indexer containing NA values"
                    )

                # Convert to unsigned integer array of indices
                indices = np.nonzero(item)[0].astype(np.uint32)
                if len(indices):
                    return self.take(indices, allow_fill=False)
                else:
                    return self[:0]
            elif kind in ('i', 'u'):
                if any(pd.isna(item)):
                    raise ValueError(
                        "Cannot index with an integer indexer containing NA values"
                    )
                return self.take(item, allow_fill=False)
            else:
                raise IndexError(err_msg)
        else:
            raise IndexError(err_msg)

    def take(self, indices, allow_fill=False, fill_value=None):
        indices = np.asarray(indices)

        # Validate self non-empty (Pandas expects this error when array is empty)
        if (len(self) == 0 and len(indices) > 0 and
                (not allow_fill or any(indices >= 0))):
            raise IndexError("cannot do a non-empty take from an empty axes|out of bounds on {typ}".format(
                typ=self.__class__.__name__,
            ))

        # Validate fill values
        if allow_fill and not (
                fill_value is None or
                np.isscalar(fill_value) and np.isnan(fill_value)):

            raise ValueError('non-None fill value not supported')

        # Validate indices
        invalid_mask = indices >= len(self)
        if not allow_fill:
            invalid_mask |= indices < -len(self)

        if any(invalid_mask):
            raise IndexError(
                "Index value out of bounds for {typ} of length {n}: "
                "{idx}".format(
                    typ=self.__class__.__name__,
                    n=len(self),
                    idx=indices[invalid_mask][0]
                )
            )

        if allow_fill:
            invalid_mask = indices < -1
            if any(invalid_mask):
                # ValueError expected by pandas ExtensionArray test suite
                raise ValueError(
                    "Invalid index value for {typ} with allow_fill=True: "
                    "{idx}".format(
                        typ=self.__class__.__name__,
                        idx=indices[invalid_mask][0]
                    )
                )

            # Build pyarrow array of indices
            indices = pa.array(indices.astype('int'), mask=indices < 0)
        else:
            # Convert negative indices to positive
            negative_mask = indices < 0
            indices[negative_mask] = indices[negative_mask] + len(self)

            # Build pyarrow array of indices
            indices = pa.array(indices.astype('int'))

        return self.__class__(self.data.take(indices), dtype=self.dtype)

    take.__doc__ = ExtensionArray.take.__doc__

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=None):
        if isinstance(scalars, cls):
            return scalars
        elif isinstance(scalars, Geometry):
            scalars = [scalars]

        return cls([
            None if np.isscalar(v) and np.isnan(v) else v for v in scalars
        ], dtype=dtype)

    def _values_for_factorize(self):
        return np.array(self, dtype='object'), None

    @classmethod
    def _from_factorized(cls, values, original):
        return cls(values, dtype=original.dtype)

    def _values_for_argsort(self):
        return np.array(list(self), dtype='object')

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(
            pa.concat_arrays(
                [ea.data for ea in to_concat]
            ),
            dtype=to_concat[0].dtype
        )

    def fillna(self, value=None, method=None, limit=None):
        from pandas.api.types import is_array_like
        from pandas.core.missing import get_fill_func
        from pandas.util._validators import validate_fillna_kwargs

        value, method = validate_fillna_kwargs(value, method)

        mask = self.isna()

        if is_array_like(value):
            if len(value) != len(self):
                raise ValueError(
                    "Length of 'value' does not match. Got ({}) "
                    " expected {}".format(len(value), len(self))
                )
            value = value[mask]

        if mask.any():
            if method is not None:
                func = get_fill_func(method)
                new_values = func(self.astype(object), limit=limit, mask=mask)
                # From pandas 1.3, get_fill_func also return mask
                new_values = new_values[0] if isinstance(new_values, tuple) else new_values
                new_values = self._from_sequence(new_values, self._dtype)
            else:
                # fill with value
                new_values = np.asarray(self)
                if isinstance(value, Geometry):
                    value = [value]
                new_values[mask] = value
                new_values = self.__class__(new_values, dtype=self.dtype)
        else:
            new_values = self.copy()
        return new_values

    fillna.__doc__ = ExtensionArray.fillna.__doc__

    # Geometry properties/methods
    @property
    def sindex(self):
        if self._sindex is None:
            self.build_sindex()
        return self._sindex

    def build_sindex(self, **kwargs):
        if self._sindex is None:
            self._sindex = HilbertRtree(self.bounds, **kwargs)
        return self

    @property
    def cx(self):
        """
        Geopandas-style spatial indexer to select a subset of the array by intersection
        with a bounding box

        Format of input should be ``.cx[xmin:xmax, ymin:ymax]``. Any of
        ``xmin``, ``xmax``, ``ymin``, and ``ymax`` can be provided, but input
        must include a comma separating x and y slices. That is, ``.cx[:, :]``
        will return the full series/frame, but ``.cx[:]`` is not implemented.
        """
        return _CoordinateIndexer(self)

    @property
    def total_bounds(self):
        raise NotImplementedError

    @property
    def total_bounds_x(self):
        raise NotImplementedError

    @property
    def total_bounds_y(self):
        raise NotImplementedError

    @property
    def bounds(self):
        raise NotImplementedError

    def hilbert_distance(self, total_bounds=None, p=10):
        # Handle default total_bounds
        if total_bounds is None:
            total_bounds = list(self.total_bounds)

        # Expand zero width bounds
        if total_bounds[0] == total_bounds[2]:
            total_bounds[2] += 1.0
        if total_bounds[1] == total_bounds[3]:
            total_bounds[3] += 1.0
        total_bounds = tuple(total_bounds)

        return _distances_from_bounds(self.bounds, total_bounds, p)

    def intersects_bounds(self, bounds, inds=None):
        """
        Test whether each element in the array intersects with the supplied bounds

        Args:
            bounds: Tuple of bounds coordinates of the form (x0, y0, x1, y1)
            inds: Optional array of indices into the array. If supplied, intersection
                calculations will be performed only on the elements selected by this
                array.  If not supplied, intersection calculations are performed
                on all elements.

        Returns:
            Array of boolean values indicating which elements of the array intersect
            with the supplied bounds
        """
        raise NotImplementedError()

    def intersects(self, shape, inds=None):
        """
        Test whether each element in the array intersects with the supplied shape

        Args:
            shape: The spatialpandas shape to compute intersections with
            inds: Optional array of indices into the array. If supplied, intersection
                calculations will be performed only on the elements selected by this
                array.  If not supplied, intersection calculations are performed
                on all elements.

        Returns:
            Array of boolean values indicating which elements of the array intersect
            with the supplied shape
        """
        raise NotImplementedError(
            "intersects not yet implemented for %s objects" % type(self).__name__
        )


class _BaseCoordinateIndexer:
    def __init__(self, sindex):
        self._sindex = sindex

    def _get_bounds(self, key):
        xs, ys = key
        # Handle xs and ys as scalar numeric values
        if type(xs) is not slice:
            xs = slice(xs, xs)
        if type(ys) is not slice:
            ys = slice(ys, ys)
        if xs.step is not None or ys.step is not None:
            raise ValueError(
                "Slice step not supported. The cx indexer uses slices to represent "
                "intervals in continuous coordinate space, and a slice step has no "
                "clear interpretation in this context."
            )
        if self._sindex:
            xmin, ymin, xmax, ymax = self._sindex.total_bounds
        else:
            xmin, ymin, xmax, ymax = self._obj.total_bounds
        x0, y0, x1, y1 = (
            xs.start if xs.start is not None else xmin,
            ys.start if ys.start is not None else ymin,
            xs.stop if xs.stop is not None else xmax,
            ys.stop if ys.stop is not None else ymax,
        )
        # Handle inverted bounds
        if x1 < x0:
            x0, x1 = x1, x0
        if y1 < y0:
            y0, y1 = y1, y0
        return x0, x1, y0, y1

    def __getitem__(self, key):
        x0, x1, y0, y1 = self._get_bounds(key)
        if self._sindex:
            covers_inds, overlaps_inds = self._sindex.covers_overlaps((x0, y0, x1, y1))
        else:
            covers_inds, overlaps_inds = None, None
        return self._perform_get_item(covers_inds, overlaps_inds, x0, x1, y0, y1)

    def _perform_get_item(self, covers_inds, overlaps_inds, x0, x1, y0, y1):
        raise NotImplementedError()


class _CoordinateIndexer(_BaseCoordinateIndexer):
    def __init__(self, obj, parent=None):
        super().__init__(obj._sindex)
        self._obj = obj
        self._parent = parent

    def _perform_get_item(self, covers_inds, overlaps_inds, x0, x1, y0, y1):
        overlaps_inds_mask = self._obj.intersects_bounds(
            (x0, y0, x1, y1), overlaps_inds
        )
        if covers_inds is not None:
            selected_inds = np.sort(
                np.concatenate([covers_inds, overlaps_inds[overlaps_inds_mask]])
            )
            if self._parent is not None:
                if len(self._parent) > 0:
                    return self._parent.iloc[selected_inds]
                else:
                    return self._parent
            return self._obj[selected_inds]
        else:
            if self._parent is not None:
                if len(self._parent) > 0:
                    return self._parent[overlaps_inds_mask]
                else:
                    return self._parent
            return self._obj[overlaps_inds_mask]


@ngjit
def _perform_extract_isnull_bytemap(bitmap, bitmap_length, bitmap_offset, dst_offset, dst):
    """
    Note: Copied from fletcher: See NOTICE for license info

    (internal) write the values of a valid bitmap as bytes to a pre-allocatored
    isnull bytemap.

    Parameters
    ----------
    bitmap: pyarrow.Buffer
        bitmap where a set bit indicates that a value is valid
    bitmap_length: int
        Number of bits to read from the bitmap
    bitmap_offset: int
        Number of bits to skip from the beginning of the bitmap.
    dst_offset: int
        Number of bytes to skip from the beginning of the output
    dst: numpy.array(dtype=bool)
        Pre-allocated numpy array where a byte is set when a value is null
    """
    for i in range(bitmap_length):
        idx = bitmap_offset + i
        byte_idx = idx // 8
        bit_mask = 1 << (idx % 8)
        dst[dst_offset + i] = (bitmap[byte_idx] & bit_mask) == 0


def _extract_isnull_bytemap(list_array):
    """
    Note: Copied from fletcher: See NOTICE for license info

    Extract the valid bitmaps of a chunked array into numpy isnull bytemaps.

    Parameters
    ----------
    chunked_array: pyarrow.ChunkedArray

    Returns
    -------
    valid_bytemap: numpy.array
    """
    result = np.zeros(len(list_array), dtype=bool)

    offset = 0
    chunk = list_array
    valid_bitmap = chunk.buffers()[0]
    if valid_bitmap:
        buf = memoryview(valid_bitmap)
        _perform_extract_isnull_bytemap(buf, len(chunk), chunk.offset, offset, result)
    else:
        return np.full(len(list_array), False)

    return result


def is_geometry_array(data):
    """
    Check if the data is of geometry dtype.
    Does not include object array of GeometryList/shapely scalars
    """
    if isinstance(getattr(data, "dtype", None), GeometryDtype):
        return True
    else:
        return False


def to_geometry_array(data, dtype=None):
    from . import (LineArray, MultiLineArray, MultiPointArray,
                   MultiPolygonArray, PointArray, PolygonArray, RingArray)
    if sg is not None:
        shapely_to_spatialpandas = {
            sg.Point: PointArray,
            sg.MultiPoint: MultiPointArray,
            sg.LineString: LineArray,
            sg.LinearRing: RingArray,
            sg.MultiLineString: MultiLineArray,
            sg.Polygon: PolygonArray,
            sg.MultiPolygon: MultiPolygonArray,
        }
    else:
        shapely_to_spatialpandas = {}

    # Normalize dtype from string
    if dtype is not None:
        dtype = pd.array([], dtype=dtype).dtype

    err_msg = "Unable to convert data argument to a GeometryList array"
    if is_geometry_array(data):
        # Keep data as is
        pass
    elif (is_array_like(data) or
            isinstance(data, (list, tuple))
            or gp and isinstance(data, (gp.GeoSeries, gp.array.GeometryArray))):

        if dtype is not None:
            data = dtype.construct_array_type()(data, dtype=dtype)
        elif len(data) == 0:
            raise ValueError(
                "Cannot infer spatialpandas geometry type from empty collection "
                "without dtype.\n"
            )
        else:
            # Check for list/array of geometry scalars.
            first_valid = None
            for val in data:
                if val is not None:
                    first_valid = val
                    break
            if isinstance(first_valid, Geometry):
                # Pass data to constructor of appropriate geometry array
                data = first_valid.construct_array_type()(data)
            elif type(first_valid) in shapely_to_spatialpandas:
                if isinstance(first_valid, sg.LineString):
                    # Handle mix of sg.LineString and sg.MultiLineString
                    for val in data:
                        if isinstance(val, sg.MultiLineString):
                            first_valid = val
                            break
                elif isinstance(first_valid, sg.Polygon):
                    # Handle mix of sg.Polygon and sg.MultiPolygon
                    for val in data:
                        if isinstance(val, sg.MultiPolygon):
                            first_valid = val
                            break

                array_type = shapely_to_spatialpandas[type(first_valid)]
                data = array_type.from_geopandas(data)
            else:
                raise ValueError(err_msg)
    else:
        raise ValueError(err_msg)
    return data



from collections import OrderedDict

import dask
import dask.dataframe as dd
import geopandas as gp
import pandas as pd
import pytest
import shapely.geometry as sg

import spatialpandas as sp
from spatialpandas import GeoDataFrame, GeoSeries

dask.config.set(scheduler="single-threaded")


@pytest.mark.parametrize('use_dask', [False, True])
def test_active_geometry(use_dask):
    gdf = GeoDataFrame(OrderedDict([
        ('a', [3, 2, 1]),
        ('points', pd.array([[0, 0, 1, 1], [2, 2, 3, 3], [4, 4]], dtype='multipoint')),
        ('line', pd.array([[0, 0, 1, 1], [2, 2, 3, 3], [4, 4]], dtype='line')),
    ]))

    if use_dask:
        gdf = dd.from_pandas(gdf, npartitions=2)

    # geometry start out as first compatible column in data frame
    assert gdf.geometry.name == 'points'

    # set_geometry default to copy operation
    assert gdf.set_geometry('line').geometry.name == 'line'
    assert gdf.geometry.name == 'points'

    # set_geometry inplace mutates geometry column
    if use_dask:
        # inplace not supported for DaskGeoDataFrame
        gdf = gdf.set_geometry('line')
    else:
        gdf.set_geometry('line', inplace=True)
    assert gdf.geometry.name == 'line'

    # Activate geometry propagates through slicing
    sliced_gdf = gdf.loc[[0, 2, 1, 0]]
    assert isinstance(sliced_gdf, type(gdf))
    assert sliced_gdf.geometry.name == 'line'

    # Select columns not including active geometry
    selected_gdf = gdf[['a', 'points']]
    with pytest.raises(ValueError):
        selected_gdf.geometry

    assert selected_gdf.set_geometry('points').geometry.name == 'points'


def test_dataframe_slice_types():
    gdf = GeoDataFrame({
        'a': [3, 2, 1],
        'b': [10, 11, 12],
        'points': pd.array([[0, 0, 1, 1], [2, 2, 3, 3], [4, 4]], dtype='multipoint'),
        'line': pd.array([[0, 0, 1, 1], [2, 2, 3, 3], [4, 4]], dtype='line'),
    })

    assert isinstance(gdf['a'], pd.Series)
    assert isinstance(gdf['points'], GeoSeries)
    assert isinstance(gdf[['a', 'b']], pd.DataFrame)
    assert isinstance(gdf[['a', 'line']], GeoDataFrame)


def test_import_geopandas_preserves_geometry_column():
    gp_points = gp.array.from_shapely([
        sg.Point([0, 0]), sg.Point(1, 1)
    ])

    gp_lines = gp.array.from_shapely([
        sg.LineString([(0, 0), (1, 1)]), sg.LineString([(1, 1), (2, 2)])
    ])

    gpdf = gp.GeoDataFrame({
        'a': [1, 2], 'point': gp_points, 'b': [3, 4], 'line': gp_lines
    })

    # Import with no geometry column set
    spdf = sp.GeoDataFrame(gpdf)
    assert spdf.geometry.name == 'point'

    # Import with geometry column set
    for geom_name in ['point', 'line']:
        gpdf = gpdf.set_geometry(geom_name)
        spdf = sp.GeoDataFrame(gpdf)
        assert spdf.geometry.name == geom_name


def test_import_geopandas_with_none():
    # Construct geopandas dataframe with column for each geometry type that includes
    # Nones
    gp_points = gp.array.from_shapely([
        sg.Point([0, 0]), sg.Point(1, 1), None
    ])

    gp_lines = gp.array.from_shapely([
        sg.LineString([(0, 0), (1, 1)]), None, sg.LineString([(1, 1), (2, 2)])
    ])

    gp_rings = gp.array.from_shapely([
        sg.LineString([(0, 0), (1, 1), (0, 0)]),
        None,
        sg.LineString([(1, 1), (2, 2), (1, 1)]),
    ])

    gp_multilines = gp.array.from_shapely([
        None,
        sg.MultiLineString([[(0, 0), (1, 1)]]),
        sg.MultiLineString([[(1, 1), (2, 2)]])
    ])

    gp_polygons = gp.array.from_shapely([
        None,
        sg.Polygon([(0, 0), (1, 1), (0, 1)]),
        sg.Polygon([(1, 1), (2, 2), (1, 0)])
    ])

    gp_multipolygons = gp.array.from_shapely([
        sg.MultiPolygon([sg.Polygon([(0, 0), (1, 1), (0, 1)])]),
        sg.MultiPolygon([sg.Polygon([(1, 1), (2, 2), (1, 0)])]),
        None
    ])

    gpdf = gp.GeoDataFrame({
        'a': [1, 2, 3],
        'point': gp_points,
        'b': [3, 4, 5],
        'line': gp_lines,
        'ring': gp_rings,
        'multiline': gp_multilines,
        'polygon': gp_polygons,
        'multipolygon': gp_multipolygons,

    })

    # Construct spatialpandas GeoDataFrame
    spdf = sp.GeoDataFrame(gpdf)

    # Check that expected entries are NA
    pd.testing.assert_frame_equal(
        spdf.isna(),
        pd.DataFrame({
            'a': [0, 0, 0],
            'point': [0, 0, 1],
            'b': [0, 0, 0],
            'line': [0, 1, 0],
            'ring': [0, 1, 0],
            'multiline': [1, 0, 0],
            'polygon': [1, 0, 0],
            'multipolygon': [0, 0, 1]
        }, dtype='bool')
    )


def test_drop_geometry_column():
    gp_points = gp.array.from_shapely([
        sg.Point([0, 0]), sg.Point(1, 1)
    ])

    gp_lines = gp.array.from_shapely([
        sg.LineString([(0, 0), (1, 1)]), sg.LineString([(1, 1), (2, 2)])
    ])

    gpdf = gp.GeoDataFrame({
        'a': [1, 2],
        'point': gp_points,
        'b': [3, 4],
        'line1': gp_lines,
        'line2': gp_lines
    }, geometry='line2')

    # Import with no geometry column set
    spdf = sp.GeoDataFrame(gpdf)
    assert spdf.geometry.name == 'line2'

    # Active geometry column preserved when dropping a different non-geometry column
    df = spdf.drop('b', axis=1)
    assert df.geometry.name == 'line2'

    # Active geometry column preserved when dropping a different geometry column
    df = spdf.drop('line1', axis=1)
    assert df.geometry.name == 'line2'

    # Dropping active geometry column results in GeoDataFrame with no active geometry
    df = spdf.drop(['point', 'line2'], axis=1)
    assert df._has_valid_geometry() is False

    # Dropping all geometry columns results in pandas DataFrame
    pdf = spdf.drop(['point', 'line1', 'line2'], axis=1)
    assert type(pdf) is pd.DataFrame



import pandas as pd

from ._optional_imports import gp
from .geometry import GeometryDtype
from .geoseries import GeoSeries, _MaybeGeoSeries


class _MaybeGeoDataFrame(pd.DataFrame):
    def __new__(cls, *args, **kwargs):
        try:
            return GeoDataFrame(*args, **kwargs)
        except ValueError:
            # No geometry compatible columns
            return pd.DataFrame(*args, **kwargs)


class GeoDataFrame(pd.DataFrame):
    # properties to propagate
    _metadata = ['_geometry']

    def __init__(self, data=None, index=None, geometry=None, **kwargs):
        # Call pandas constructor, always copy
        kwargs.pop("copy", None)
        super().__init__(data, index=index, copy=True, **kwargs)

        # Replace pd.Series of GeometryArrays with GeoSeries.
        first_geometry_col = None
        for col in self.columns:
            if (isinstance(self[col].dtype, GeometryDtype) or
                    gp and isinstance(self[col].dtype, gp.array.GeometryDtype)):
                self[col] = GeoSeries(self[col])
                first_geometry_col = first_geometry_col or col

        if first_geometry_col is None:
            raise ValueError(
                "A spatialpandas GeoDataFrame must contain at least one spatialpandas "
                "GeometryArray column"
            )

        if geometry is None:
            if isinstance(data, GeoDataFrame) and data._has_valid_geometry():
                geometry = data._geometry
            elif gp and isinstance(data, gp.GeoDataFrame):
                try:
                    geometry = data.geometry.name
                except AttributeError:
                    # Geometry column not set
                    pass

        if geometry is None:
            geometry = first_geometry_col

        self._geometry = None
        if geometry is not None:
            self.set_geometry(geometry, inplace=True)

    @property
    def _constructor(self):
        return _MaybeGeoDataFrame

    @property
    def _constructor_sliced(self):
        return _MaybeGeoSeries

    def set_geometry(self, geometry, inplace=False):
        if (geometry not in self or
                not isinstance(self[geometry].dtype, GeometryDtype)):
            raise ValueError(
                "The geometry argument must be the name of a spatialpandas "
                "geometry column in the spatialpandas GeoDataFrame"
            )

        if inplace:
            self._geometry = geometry
            return self
        else:
            return GeoDataFrame(self, geometry=geometry)

    def _has_valid_geometry(self):
        if (self._geometry is not None and
                self._geometry in self and
                isinstance(self[self._geometry].dtype, GeometryDtype)):
            return True
        else:
            return False

    @property
    def geometry(self):
        if not self._has_valid_geometry():
            raise ValueError(
                "GeoDataFrame has no active geometry column.\n"
                "The active geometry column should be set using the set_geometry "
                "method."
            )

        return self[self._geometry]

    def to_geopandas(self):
        from geopandas import GeoDataFrame as gp_GeoDataFrame
        gdf = gp_GeoDataFrame(
            {col: s.array.to_geopandas() if isinstance(s.dtype, GeometryDtype) else s
             for col, s in self.items()},
            index=self.index,
        )
        if self._has_valid_geometry():
            gdf.set_geometry(self._geometry, inplace=True)
        return gdf

    @property
    def cx(self):
        from .geometry.base import _CoordinateIndexer
        return _CoordinateIndexer(
            self.geometry.array, parent=self
        )

    def build_sindex(self, **kwargs):
        self.geometry.build_sindex(**kwargs)
        return self

    def _ensure_type(self, obj):
        # Override because a GeoDataFrame operation may result in a regular DataFrame,
        # and that's ok
        assert isinstance(obj, type(self)) or isinstance(self, type(obj)), type(obj)
        return obj



import numpy as np
import pandas as pd
from dask import delayed
from dask.dataframe import from_delayed, from_pandas


def _record_reset_index(df, suffix):
    # Record original index name(s), generate new index name(s), reset index column(s)
    new_column_name = ["index_%s" % suffix]
    df = df.copy(deep=True)
    try:
        old_index_name = [df.index.name]
        df.index = df.index.rename(new_column_name[0])
    except TypeError:
        new_column_name = [
            "index_%s" % suffix + str(l) for l, ix in
            enumerate(df.index.names)
        ]
        old_index_name = df.index.names
        df.index = df.index.rename(new_column_name)
    df = df.reset_index(drop=False)

    return df, old_index_name, new_column_name


def sjoin(
        left_df, right_df, how="inner", op="intersects",
        lsuffix="left", rsuffix="right"
):
    """
    Spatial join between two GeoDataFrames or between a DaskGeoDataFrame and
    a GeoDataFrame

    Args:
        left_df: A GeoDataFrame or DaskGeoDataFrame
        right_df: A GeoDataFrame
        how: The type of join. One of 'inner', 'left', or 'right'. Note that 'right'
            is not supported when left_df is a DaskGeoDataFrame
        op: Binary predicate, currently only "intersects" is supported
        lsuffix: Suffix to apply to overlapping column names from the left GeoDataFrame
        rsuffix: Suffix to apply to overlapping column names from the right GeoDataFrame

    Returns:
        GeoDataFrame or DaskGeoDataFrame (same type as left_df argument)
    """
    from .. import GeoDataFrame
    try:
        from ..dask import DaskGeoDataFrame
    except ImportError:
        DaskGeoDataFrame = type(None)

    # Validate data frame types
    if not isinstance(left_df, (GeoDataFrame, DaskGeoDataFrame)):
        raise ValueError(
            "`left_df` must be a spatialpandas.GeoDataFrame or "
            "spatialpandas.dask.DaskGeoDataFrame\n"
            "    Received value of type: {typ}".format(typ=type(left_df)))

    if not isinstance(right_df, GeoDataFrame):
        raise ValueError(
            "`right_df` must be a spatialpandas.GeoDataFrame\n"
            "    Received value of type: {typ}".format(typ=type(right_df)))

    # Validate op
    valid_op = ["intersects"]
    if op not in valid_op:
        raise ValueError(
            "`op` must be one of {valid_op}\n"
            "    Received: {val}".format(val=repr(how), valid_op=valid_op)
        )

    # Validate join type
    valid_how = ["left", "right", "inner"]
    if how not in valid_how:
        raise ValueError(
            "`how` must be one of {valid_how}\n"
            "    Received: {val}".format(val=repr(how), valid_how=valid_how)
        )

    # Validate suffixes
    if lsuffix == rsuffix:
        raise ValueError(
            "`lsuffix` and `rsuffix` must not be equal"
        )

    # Perform sjoin
    if isinstance(left_df, GeoDataFrame):
        return _sjoin_pandas_pandas(
            left_df, right_df, how=how, op=op, lsuffix=lsuffix, rsuffix=rsuffix
        )
    elif isinstance(left_df, DaskGeoDataFrame):
        return _sjoin_dask_pandas(
            left_df, right_df, how=how, op=op, lsuffix=lsuffix, rsuffix=rsuffix
        )


def _sjoin_dask_pandas(
        left_ddf, right_df, how="inner", op="intersects",
        lsuffix="left", rsuffix="right"
):
    if how == "right":
        raise ValueError(
            "`how` may not be 'right' when left_df is a DaskGeoDataFrame"
        )
    dfs = left_ddf.to_delayed()
    partition_bounds = left_ddf.geometry.partition_bounds
    sjoin_pandas = delayed(_sjoin_pandas_pandas)

    # Get spatial index for right_df
    right_sindex = right_df.geometry.sindex

    # Build list of delayed sjoin results
    joined_dfs = []
    for df, (i, bounds) in zip(dfs, partition_bounds.iterrows()):
        right_inds = right_sindex.intersects(bounds.values)
        if how == "left" or len(right_inds) > 0:
            joined_dfs.append(
                sjoin_pandas(
                    df, right_df.iloc[right_inds], how=how,
                    rsuffix=rsuffix, lsuffix=lsuffix
                )
            )

    # Compute meta
    meta = _sjoin_pandas_pandas(
        left_ddf._meta, right_df.iloc[:0], how=how, lsuffix=lsuffix, rsuffix=rsuffix
    )

    # Build resulting Dask DataFrame
    if not joined_dfs:
        return from_pandas(meta, npartitions=1)
    else:
        return from_delayed(joined_dfs, meta=meta)


def _sjoin_pandas_pandas(
        left_df, right_df, how="inner", op="intersects",
        lsuffix="left", rsuffix="right"
):
    from .. import GeoDataFrame

    # Record original index name(s), generate new index name(s), reset index column(s)
    original_right_df = right_df
    original_left_df = left_df
    right_df, right_index_name, index_right = _record_reset_index(
        original_right_df, rsuffix
    )
    left_df, left_index_name, index_left = _record_reset_index(
        original_left_df, lsuffix
    )

    if any(original_left_df.columns.isin(index_left + index_right)) or any(
            original_right_df.columns.isin(index_left + index_right)
    ):
        raise ValueError(
            "'{0}' and '{1}' cannot be column names in the GeoDataFrames being"
            " joined".format(index_left, index_right)
        )

    # Get spatial index for left frame
    sindex = left_df.geometry.sindex
    left_geom = left_df.geometry.array
    right_geom = right_df.geometry.array

    # Get bounds from right geometry
    right_bounds = right_df.geometry.bounds.values

    # Init list of arrays, the same length as right_df, where each array holds the
    # indices into left_df that intersect with the corresponding element.
    left_inds = [np.array([], dtype='uint32')] * len(right_df)

    # right_inds will hold the inds into right_df that correspond to left_inds
    right_inds = [np.array([], dtype='uint32')] * len(right_df)

    # Loop over the right frame
    for i in range(len(right_df)):
        # Get bounds for shape in current row of right_df
        shape_bounds = right_bounds[i, :]

        # Use spatial index on left_df to get indices of shapes with bounding boxes that
        # intersect with these bounds
        candidate_inds = sindex.intersects(shape_bounds)

        if len(candidate_inds) > 0:
            right_shape = right_geom[i]
            intersecting_mask = left_geom.intersects(right_shape, inds=candidate_inds)
            intersecting_inds = candidate_inds[intersecting_mask]
            left_inds[i] = intersecting_inds
            right_inds[i] = np.full(len(intersecting_inds), i)

    # Flatten nested arrays of indices
    if left_inds:
        flat_left_inds = np.concatenate(left_inds)
        flat_right_inds = np.concatenate(right_inds)
    else:
        flat_left_inds = np.array([], dtype='uint32')
        flat_right_inds = np.array([], dtype='uint32')

    # Build pandas DataFrame from inds
    result = pd.DataFrame({
        '_key_left': flat_left_inds,
        '_key_right': flat_right_inds
    })

    # Perform join
    if how == "inner":
        result = result.set_index("_key_left")
        joined = (
            left_df.merge(
                result, left_index=True, right_index=True
            ).merge(
                right_df.drop(right_df.geometry.name, axis=1),
                left_on="_key_right",
                right_index=True,
                suffixes=("_%s" % lsuffix, "_%s" % rsuffix),
            ).set_index(
                index_left
            ).drop(
                ["_key_right"], axis=1
            )
        )
        if len(left_index_name) > 1:
            joined.index.names = left_index_name
        else:
            joined.index.name = left_index_name[0]

    elif how == "left":
        result = result.set_index("_key_left")
        joined = (
            left_df.merge(
                result, left_index=True, right_index=True, how="left"
            ).merge(
                right_df.drop(right_df.geometry.name, axis=1),
                how="left",
                left_on="_key_right",
                right_index=True,
                suffixes=("_%s" % lsuffix, "_%s" % rsuffix),
            ).set_index(
                index_left
            ).drop(
                ["_key_right"], axis=1
            )
        )
        if len(left_index_name) > 1:
            joined.index.names = left_index_name
        else:
            joined.index.name = left_index_name[0]

    else:  # how == 'right':
        joined = (
            left_df.drop(
                left_df.geometry.name, axis=1
            ).merge(
                result.merge(
                    right_df, left_on="_key_right", right_index=True, how="right"
                ),
                left_index=True,
                right_on="_key_left",
                suffixes=("_%s" % lsuffix, "_%s" % rsuffix),
                how="right",
            ).set_index(
                index_right
            ).drop(
                ["_key_left", "_key_right"], axis=1
            )
        )
        if len(right_index_name) > 1:
            joined.index.names = right_index_name
        else:
            joined.index.name = right_index_name[0]

    return GeoDataFrame(joined)
